// Mpq.cpp : Defines the entry point for the console application.
//

#include "stdafx.h"
#include "lmpqapi.h"

void AddFileToArchive(LPCSTR arcname,LPCSTR filename,LPCSTR intname,DWORD CompressMethod);
void ListArchive(LPCSTR arcname,LPCSTR files[],DWORD filcount);
void ExtactFile(LPCSTR arcname,LPCSTR intname,LPCSTR Dir,bool XCmd);
void DeleteFileFromArchive(LPCSTR arcname,LPCSTR filename);
void RenameFileInArchive(LPCSTR arcname,LPCSTR oldfilename,LPCSTR newfilename);
void CompactArchive(LPCSTR arcname);

LPCSTR UseBanner=
//"\n"
"                Usage:  mpq <command> <archive_name> <file_names>\n"
"  Add or replace file:  mpq a <archive_name> <file_name> <packed_name>\n"
"     Add uncompressed:  mpq A <archive_name> <file_name> <packed_name>\n"
"          Delete file:  mpq d <archive_name> <file_name>\n"
"          Rename file:  mpq r <archive_name> <file_name> <new_name>\n"
"      Compact archive:  mpq c <archive_name>\n"
"         Extract file:  mpq e <archive_name> <packed_name> [target_directory]\n"
"Extract with pathname:  mpq x <archive_name> <packed_name> [target_directory]\n"
"        List archive :  mpq l <archive_name> [listfile1] [listfile2] [...]\n"
"";

/////////////////////////////////////////////////////////////////////////////
// The main program procedure

int main(int argc, char* argv[])
{
  if (MpqInitialize()==FALSE) {
    cputs("Can't initialize LMPQAPI.dll\n");
    return 2;
  };

	printf("LMPQDLL demo program v1.2 Copyright (c) 1999 Andrey V. Lelikov\n");
  cputs("This programs uses "); cputs(MpqGetVersionString()); cputs("\n\n");

	if (argc < 3)
	{
    banner:
		cputs(UseBanner);
		return 1;
	}
  switch(*argv[1]) {
  case 'a' :
    AddFileToArchive(argv[2],argv[3],(argc>=5)?argv[4]:0,1);
    break;
  case 'A' :
    AddFileToArchive(argv[2],argv[3],(argc>=5)?argv[4]:0,0);
    break;
  case 'd' :
  case 'D' :
    DeleteFileFromArchive(argv[2],argv[3]);
    break;
  case 'r' :
  case 'R' :
    RenameFileInArchive(argv[2],argv[3],argv[4]);
    break;
  case 'c' :
  case 'C' :
    CompactArchive(argv[2]);
	break;
  case 'l':
  case 'L':
    ListArchive(argv[2],(LPCSTR*)argv+3,argc-3);
    break;
  case 'e':
  case 'E':
    ExtactFile(argv[2],(argc>=4)?argv[3]:0,(argc>=5)?argv[4]:0,false);
    break;
  case 'X':
  case 'x':
    ExtactFile(argv[2],(argc>=4)?argv[3]:0,(argc>=5)?argv[4]:0,true);
    break;
  default:
    goto banner;
  };
  return 0;
}

void DeleteFileFromArchive(LPCSTR arcname,LPCSTR filename)
{
  MPQHANDLE mpq;
  if (filename==0) {
    cputs("Must specify file to delete.\n");
	return;
  };
  mpq=MpqOpenArchiveForUpdate(arcname,MOAU_OPEN_EXISTING|MOAU_MAINTAIN_LISTFILE,1024); // open
  if ( (mpq==0) || (mpq==((MPQHANDLE)0xffffffff)) ) {
    cputs("Error: Can't open archive ");
    cputs(arcname);
    cputs("\n");
    return;
  };
  cputs("Archive: "); cputs(arcname); cputs("\n");
  BOOL rtn;
  rtn=MpqDeleteFile(mpq,filename);
  if (rtn==0) {
    cputs("Error: Can't delete file ");
    cputs(filename);
    cputs("\n");
    MpqCloseUpdatedArchive(mpq,0);
    return;
  }; 
  cputs("Deleting file ");
  cputs(filename);
  cputs("... Ok.\n");
  MpqCloseUpdatedArchive(mpq,0);
};

void RenameFileInArchive(LPCSTR arcname,LPCSTR oldfilename,LPCSTR newfilename)
{
  MPQHANDLE mpq;
  if (oldfilename==0 || newfilename==0) {
    cputs("Must specify old name and new name.\n");
	return;
  };
  mpq=MpqOpenArchiveForUpdate(arcname,MOAU_OPEN_EXISTING|MOAU_MAINTAIN_LISTFILE,1024); // open
  if ( (mpq==0) || (mpq==((MPQHANDLE)0xffffffff)) ) {
    cputs("Error: Can't open archive ");
    cputs(arcname);
    cputs("\n");
    return;
  };
  cputs("Archive: "); cputs(arcname); cputs("\n");
  BOOL rtn;
  rtn=MpqRenameFile(mpq,oldfilename,newfilename);
  if (rtn==0) {
    cputs("Error: Can't rename file ");
    cputs(oldfilename);
    cputs("\n");
    MpqCloseUpdatedArchive(mpq,0);
    return;
  }; 
  cputs("Renaming file ");
  cputs(oldfilename);
  cputs(" => ");
  cputs(newfilename);
  cputs("... Ok.\n");
  MpqCloseUpdatedArchive(mpq,0);
};

void CompactArchive(LPCSTR arcname)
{
  MPQHANDLE mpq;
  mpq=MpqOpenArchiveForUpdate(arcname,MOAU_OPEN_EXISTING|MOAU_MAINTAIN_LISTFILE,1024); // open
  if ( (mpq==0) || (mpq==((MPQHANDLE)0xffffffff)) ) {
    cputs("Error: Can't open archive ");
    cputs(arcname);
    cputs("\n");
    return;
  };
  cputs("Archive: "); cputs(arcname); cputs("\n");
  cputs("Compacting...");
  BOOL rtn;
  rtn=MpqCompactArchive(mpq);
  if (rtn==0) {
    cputs("\nError: Can't compact archive\n");
    MpqCloseUpdatedArchive(mpq,0);
    return;
  }; 
  cputs(" Ok.\n");
  MpqCloseUpdatedArchive(mpq,0);
}

void AddFileToArchive(LPCSTR arcname,LPCSTR filename,LPCSTR intname,DWORD CompressMethod)
{
  MPQHANDLE mpq;
  if (filename==0) {
    cputs("Must specify what file to add.\n");
	return;
  };
  mpq=MpqOpenArchiveForUpdate(arcname,MOAU_OPEN_EXISTING|MOAU_MAINTAIN_LISTFILE,1024); // open
  if ((mpq==0) || (mpq==((MPQHANDLE)0xffffffff)) ) {
    mpq=MpqOpenArchiveForUpdate(arcname,MOAU_CREATE_NEW|MOAU_MAINTAIN_LISTFILE,1024); // create
  };
  if ( (mpq==0) || (mpq==((MPQHANDLE)0xffffffff)) ) {
    cputs("Error: Can't create archive ");
    cputs(arcname);
    cputs("\n");
    return;
  };
  cputs("Archive: "); cputs(arcname); cputs("\n");

  DWORD aflgs;
  aflgs=(CompressMethod==0)?MAFA_REPLACE_EXISTING:MAFA_COMPRESS|MAFA_ENCRYPT|MAFA_REPLACE_EXISTING;
  if (intname==0) {
    intname=filename;
    char *tc;
    tc=strchr(intname,':');
    if (tc!=0) { 
      intname=tc+1; 
      if (*intname=='\\') {
        intname++;
      };
    };
  };
  cputs("Adding file ");
  cputs(intname);
  cputs(" ...");
  MpqAddFileToArchive(mpq,filename,intname,aflgs);
  MpqCloseUpdatedArchive(mpq,0);
  cputs(" Ok\n");
};

LPVOID MapMpqFile(MPQHANDLE mpq,LPCSTR Name,DWORD *psz)
{
  MPQHANDLE file;
  if (SFileOpenFileEx(mpq,Name,0,&file)==0) return(0);
  DWORD sz,ss;
  sz=SFileGetFileSize(file,0);
  if (psz!=0) *psz=sz;
  if (sz==0) return(0);
  LPVOID rtn;
  rtn=VirtualAlloc(0,sz,MEM_COMMIT,PAGE_READWRITE);
  if (rtn==0) return(0);
  ss=SFileReadFile(file,rtn,sz,0,0);
  SFileCloseFile(file);
  if (ss==0) {
    VirtualFree(rtn,0,MEM_RELEASE);
    return(0);
  };
  return(rtn);
};

void UnmapMpqFile(LPVOID file)
{
  VirtualFree(file,0,MEM_RELEASE);
};

LPVOID MapOrdinalFile(LPCSTR Name,DWORD *psz)
{
  HANDLE file;
  file=CreateFile(Name,GENERIC_READ,FILE_SHARE_READ,0,OPEN_EXISTING,0,0);
  if (file==INVALID_HANDLE_VALUE) return(0);
  DWORD sz,ss,tempsz;
  sz=GetFileSize(file,0);
  if (psz!=0) *psz=sz;
  if (sz==0) {CloseHandle(file);return(0);}
  LPVOID rtn;
  rtn=VirtualAlloc(0,sz,MEM_COMMIT,PAGE_READWRITE);
  if (rtn==0) {CloseHandle(file);return(0);}
  ss=ReadFile(file,rtn,sz,&tempsz,0);
  DWORD ErrorValue = GetLastError();
  CloseHandle(file);
  if (ss==0) {
    VirtualFree(rtn,0,MEM_RELEASE);
    return(0);
  };
  return(rtn);
};

void UnmapOrdinalFile(LPVOID file)
{
  VirtualFree(file,0,MEM_RELEASE);
};

#include <pshpack1.h>

class LineItem {
public:
  LPCSTR p;
  USHORT sz;
  static int __cdecl compare(const void *elem1, const void *elem2);
};


typedef enum _MapMethod {
  MapNone,
  MapMpq,
  MapFile
} MapMethod;

class MappedItem {
public:
  MapMethod mm;
  LPVOID m;
  DWORD sz;
  LPCSTR GetEol(LPCSTR Ptr);
  LPCSTR GetNextLine(LPCSTR Ptr);
  void Unmap();
};

#include <poppack.h>

int __cdecl LineItem::compare(const void *elem1, const void *elem2)
{
  char *l1,*l2;
  LineItem *e1,*e2;
  e1=(LineItem *)elem1;
  e2=(LineItem *)elem2;
  l1=(char*)_alloca(e1->sz+1);
  l2=(char*)_alloca(e2->sz+1);
  memcpy(l1,e1->p,e1->sz); l1[e1->sz]=0;
  memcpy(l2,e2->p,e2->sz); l2[e2->sz]=0;
  return(stricmp(l1,l2));
};

LPCSTR MappedItem::GetEol(LPCSTR Ptr)
{
  LPCSTR tt=Ptr;
  while (*tt!='\n') {
    if (tt==( ((LPCSTR)m) + sz ) ) return(tt);
    tt++;
  };
  for(;tt!=Ptr;tt--) {
    if (*tt=='\n') continue;
    if (*tt=='\r') continue;
    break;
  };
  return(tt+1);
};

LPCSTR MappedItem::GetNextLine(LPCSTR Ptr)
{
  LPCSTR tt=Ptr;
  while (*tt!='\n') {
    if (tt==( ((LPCSTR)m) + sz ) ) return(0);
    tt++;
  };
  tt++;
  if (*tt=='\r') tt++;
  return(tt);
};

void MappedItem::Unmap()
{
  switch(this->mm) {
    case MapMpq: UnmapMpqFile(m); break;
    case MapFile: UnmapOrdinalFile(m); break;
  };
};

void ListArchive(LPCSTR arcname,LPCSTR files[],DWORD filcount)
{
  MPQHANDLE mpq;
  if (SFileOpenArchive(arcname,0,0,&mpq)==0) {
    cputs("Error: can't open archive ");
    cputs(arcname);
    cputs("\n");
    return;
  };

  cputs("Creating listfile...");
  MappedItem *fls;
  fls=(MappedItem *)_alloca(sizeof(MappedItem)*(filcount+2));
//  memset(fls,0,sizeof(MappedItem)*(filcount+2));
  fls[0].m="(listfile)\n(signature)\n";
  fls[0].sz=strlen((LPCSTR)fls[0].m)+1;
  fls[0].mm=MapNone;
  fls[1].m=MapMpqFile(mpq,"(listfile)",&fls[1].sz);
  fls[1].mm=MapMpq;

  DWORD rfc;
  rfc=2;
  for (unsigned int i=0,j=0;i<filcount;i++) {
    fls[j+2].m=MapOrdinalFile(files[i],&fls[j+2].sz);
    if (fls[j+2].m==0) continue;
    j++;
    rfc++;
  };

  // counting lines
  DWORD lcnt=0;
  for (i=0;i<rfc;i++) {
    if (fls[i].m==0) continue;
    LPCSTR ts;
    ts=(LPCSTR)fls[i].m;
    do {
      lcnt++;
      ts=fls[i].GetNextLine(ts);
    } while(ts!=0);
  };

  // allocating lines
  LineItem *Lines;
  Lines=(LineItem *)VirtualAlloc(0,sizeof(LineItem)*lcnt,MEM_COMMIT,PAGE_READWRITE);
  if (Lines==0) {
    cputs("\nError: No memory\n");
    return;
  };
  j=0;
  for (i=0;i<rfc;i++) {
    LPCSTR ts;
    ts=(LPCSTR)fls[i].m;
    if (fls[i].m==0) continue;
    do {
      Lines[j].p=ts;
      Lines[j].sz=fls[i].GetEol(ts)-ts;
      ts=fls[i].GetNextLine(ts);
      if (Lines[j].sz!=0) j++;
      if (ts==0) break;
    } while(true);
  };

  // sorting array
  qsort(Lines,j,sizeof(LineItem),LineItem::compare);

  // deleting duplicates
  DWORD LargestStr=0;
  for (i=0;i<(j-1);i++) {
    if (Lines[i].sz>LargestStr) LargestStr=Lines[i].sz;
    if (Lines[i+1].sz>LargestStr) LargestStr=Lines[i+1].sz;
    if (LineItem::compare(&Lines[i],&Lines[i+1])!=0) continue;
    j--;
    memmove(&Lines[i],&Lines[i+1],(j-i)*sizeof(LineItem));
    i=0-1;
  };

  // ok, name array allocated, let's list it!!!
  char *tstr;
  tstr=(char*)_alloca(LargestStr+2);
  cputs(" Done\n\n");
  cputs("Archive: "); cputs(arcname); cputs("\n");
  cputs("Filename                                Size\n");
  cputs("--------------------------------------  ------------\n");
  DWORD mfilcnt=0,mallsize=0;
  for (i=0;i<j;i++) {
    memcpy(tstr,Lines[i].p,Lines[i].sz);
    tstr[Lines[i].sz]=0;
    MPQHANDLE tfile;
    if (SFileOpenFileEx(mpq,tstr,0,&tfile)==0) continue;
    cputs(tstr);
    if (strlen(tstr)<40) {
      int pl;
      pl=40-strlen(tstr);
      for (int k=0;k<pl;k++) cputs(" ");
    };
    DWORD fsz;
    fsz=SFileGetFileSize(tfile,0);
    mallsize+=fsz;
    printf("%12d\n",fsz);
    SFileCloseFile(tfile);
    mfilcnt++;
  };
  cputs("--------------------------------------  ------------\n");
  printf("%5d recognized files                  %12d\n",mfilcnt,mallsize);

  // unmapping files
  for (i=0;i<rfc;i++) {
    if (fls[i].m==0) continue;
    fls[i].Unmap();
  };
  SFileCloseArchive(mpq);
};

void ExtactFile(LPCSTR arcname,LPCSTR intname,LPCSTR Dir,bool XCmd)
{
  MPQHANDLE mpq;
  if (intname==0) {
    cputs("Invalid filename\n");
    return;
  };
  if (SFileOpenArchive(arcname,0,0,&mpq)==0) {
    cputs("Error: can't open archive ");
    cputs(arcname);
    cputs("\n");
    return;
  };
  cputs("Archive: "); cputs(arcname); cputs("\n");

  LPVOID mpf;
  DWORD fz;
  mpf=MapMpqFile(mpq,intname,&fz);
  if (mpf==0) {
    cputs("Error: no file - "); cputs(intname); cputs("\n");
    SFileCloseArchive(mpq);
    return;
  };

  if (Dir==0) {
    Dir=".\\";
  };
  char *tp,*tn,*tpi;
  const char *ti;
  DWORD ln;
  ln=strlen(Dir) + strlen(intname) + 10;
  tn=(char*)_alloca(ln);
  strcpy(tn,Dir);
  tp=tn+strlen(tn);
  if (*(tp-1)!='\\') {
    *tp='\\';
    tp++;
  };
  if (XCmd==true) {
    ti=intname;
  } else {
    for(tpi=(char*)intname+strlen(intname);tpi!=intname;tpi--)
      if ( *(tpi-1)== '\\' ) break;
    ti=tpi;
  };

  strcpy(tp,ti);

  cputs(intname); cputs(" => "); cputs(tn); 

  HANDLE of;

  char *tnp;
  tnp=(char*)_alloca(strlen(tn)+1);
  lstrcpy(tnp,tn);
  DWORD tnplen = strlen(tnp);
  for(tpi=tnp;tpi<=tnp+tnplen;tpi++)
    if (*(tpi)== '\\') {
        *tpi=0;
        CreateDirectory(tnp,0);
        memcpy(tpi,"\\",1);
	}

  of=CreateFile(tn,GENERIC_READ|GENERIC_WRITE,0,0,CREATE_ALWAYS,0,0);
  if (of==INVALID_HANDLE_VALUE) {
    cputs("Error: can't create file "); cputs(tp); cputs("\n");
    UnmapMpqFile(mpf);
    SFileCloseArchive(mpq);
  };
  DWORD ff;
  WriteFile(of,mpf,fz,&ff,0);
  CloseHandle(of);

  cputs("\n");
  UnmapMpqFile(mpf);

  SFileCloseArchive(mpq);
};

